
from random import random
from threading import Thread

import pytest

from wasp_c_extensions.ev_loop import WEventLoop


class TestEventLoopConcurrency:
	__test_running__ = False
	__threads_count__ = 50
	__input_sequence__ = [int(random() * 10) for _ in range(10 ** 4)]
	__wait_pause__ = 0.01

	def test(self):
		loop = WEventLoop()
		thread = Thread(target=loop.start_loop)
		thread.start()
		loop.stop_loop()
		thread.join()

	def test_exception(self):
		def callback():
			raise ValueError('!')

		loop = WEventLoop(immediate_stop=False)
		loop.notify(callback)

		pytest.raises(ValueError, loop.start_loop)

	@pytest.mark.parametrize('runs', range(5))
	def test_concurrency(self, runs):
		loop = WEventLoop(immediate_stop=False)
		loop_thread = Thread(target=loop.start_loop)
		loop_thread.start()

		result = []

		class Callback:
			def __init__(self, i):
				self.i = i

			def __call__(self, *args, **kwargs):
				result.append(self.i)

		def pub_thread_fn():
			for i in TestEventLoopConcurrency.__input_sequence__:
				loop.notify(Callback(i))

		pub_threads = [Thread(target=pub_thread_fn) for _ in range(TestEventLoopConcurrency.__threads_count__)]
		for i in pub_threads:
			i.start()

		for i in pub_threads:
			i.join()

		loop.stop_loop()
		loop_thread.join()

		target_result = sum(TestEventLoopConcurrency.__input_sequence__) * TestEventLoopConcurrency.__threads_count__
		assert(sum(result) == target_result)




import pytest
import threading

from wasp_c_extensions.threads import WAtomicCounter, WPThreadEvent, awareness_wait
from wasp_c_extensions.threads import LT_test, LE_test, EQ_test, NE_test, GT_test, GE_test


class TestWAtomicCounter:

	__threads__ = 50
	__repeats__ = 50

	def test(self):

		c = WAtomicCounter()
		assert(c.__int__() == 0)

		c = WAtomicCounter()
		assert(c.__int__() == 0)

		c.increase_counter(1)
		assert(c.__int__() == 1)

		c.increase_counter(10)
		assert(c.__int__() == 11)

		c = WAtomicCounter(5)
		assert(c.__int__() == 5)

		c.increase_counter(1)
		assert(c.__int__() == 6)

	def test_counter_maximum(self):
		u_long_long_bits_count = (8 * 8)

		WAtomicCounter((1 << u_long_long_bits_count) - 1)
		WAtomicCounter((1 << u_long_long_bits_count) + 10)

	def test_multi_threading(self):
		c = WAtomicCounter()

		def thread_fn_increase():
			for i in range(self.__repeats__):
				c.increase_counter(1)

		threads = [threading.Thread(target=thread_fn_increase) for x in range(self.__threads__)]
		for th in threads:
			th.start()

		for th in threads:
			th.join()

		assert(c.__int__() == (self.__threads__ * self.__repeats__))

	def test_negative(self):
		WAtomicCounter()
		WAtomicCounter(-7)
		WAtomicCounter(-7, negative=True)
		pytest.raises(ValueError, WAtomicCounter, -7, negative=False)

		c = WAtomicCounter(negative=False)
		assert(int(WAtomicCounter()) == 0)

		c = WAtomicCounter(3, negative=False)
		assert(int(c) == 3)
		pytest.raises(ValueError, c.increase_counter, -10)
		assert(int(c) == 3)

	def test_set(self):
		counter = WAtomicCounter()
		assert(counter.set(5) == 0)
		assert(int(counter) == 5)

		counter = WAtomicCounter(1, negative=False)
		assert(counter.set(7) == 1)
		assert(int(counter) == 7)

		pytest.raises(ValueError, counter.set, -3)

	__tas_parameters__ = [
		(LT_test, 3, -1, None, 3),
		(LT_test, 3, 3, None, 3),
		(LT_test, 3, 5, 3, 5),

		(LE_test, 3, -1, None, 3),
		(LE_test, 3, 3, 3, 3),
		(LE_test, 3, 5, 3, 5),

		(EQ_test, 3, -1, None, 3),
		(EQ_test, 3, 3, 3, 3),
		(EQ_test, 3, 5, None, 3),

		(NE_test, 3, -1, 3, -1),
		(NE_test, 3, 3, None, 3),
		(NE_test, 3, 5, 3, 5),

		(GT_test, 3, -1, 3, -1),
		(GT_test, 3, 3, None, 3),
		(GT_test, 3, 5, None, 3),

		(GE_test, 3, -1, 3, -1),
		(GE_test, 3, 3, 3, 3),
		(GE_test, 3, 5, None, 3),
	]

	@pytest.mark.parametrize("test_op, start_value, compare_value, tas_result, counter_result", __tas_parameters__)
	def test_tas(self, test_op, start_value, compare_value, tas_result, counter_result):
		counter = WAtomicCounter(start_value)
		tas = counter.test_and_set(test_op, WAtomicCounter(compare_value))
		assert((tas is tas_result) or (tas == tas_result))
		assert(int(counter) == counter_result)

	@pytest.mark.parametrize("test_op", [LT_test, LE_test, EQ_test, NE_test, GT_test, GE_test])
	def test_tas_exceptions(self, test_op):
		pytest.raises(
			ValueError,
			WAtomicCounter(0, negative=False).test_and_set,
			test_op,
			WAtomicCounter(-1)
		)

	def test_cas(self):
		counter = WAtomicCounter(1)
		assert(counter.compare_and_set(WAtomicCounter(5), WAtomicCounter(7)) is None)
		assert(int(counter) == 1)

		assert(counter.compare_and_set(WAtomicCounter(1), WAtomicCounter(-7)) == 1)
		assert(int(counter) == -7)

		counter = WAtomicCounter(3, negative=False)
		pytest.raises(ValueError, counter.compare_and_set, WAtomicCounter(3), WAtomicCounter(-5))
		assert(int(counter) == 3)


class TestWPThreadEvent:

	__wait_test_timeout__ = 3

	__threads__ = 50
	__repeats__ = 50

	def test(self):
		event = WPThreadEvent()
		assert(event.is_set() is False)
		event.clear()
		assert(event.is_set() is False)

		event.set()
		assert(event.is_set() is True)

		event = WPThreadEvent(None)
		assert(event.is_set() is False)
		event.clear()
		assert(event.is_set() is False)

		event.set()
		assert(event.is_set() is True)

		assert(event.wait() is True)
		assert(event.wait(None) is True)

		event.set()
		assert(event.is_set() is True)

		event.clear()
		assert(event.is_set() is False)

		assert(event.wait(self.__wait_test_timeout__) is False)
		assert(event.is_set() is False)

	def test_concurrency(self):
		event = WPThreadEvent()

		def threading_fn():
			event.wait()

		threads = [threading.Thread(target=threading_fn) for _ in range(10)]

		for th in threads:
			th.start()

		event.set()

		for th in threads:
			th.join()

	def test_multi_threading(self):

		self.test_counter = 0

		events = [WPThreadEvent() for _ in range(self.__threads__)]

		def threading_fn_gen(wait_event_obj, set_event_obj):
			def threading_fn():
				for _ in range(self.__repeats__):
					wait_event_obj.wait()
					wait_event_obj.clear()
					self.test_counter += 1
					set_event_obj.set()
			return threading_fn

		threads = [
			threading.Thread(target=threading_fn_gen(
				events[i], events[(i + 1) % self.__threads__]
			)) for i in range(self.__threads__)
		]

		for th in threads:
			th.start()

		events[0].set()

		for th in threads:
			th.join()

		assert(self.test_counter == (self.__threads__ * self.__repeats__))


@pytest.mark.parametrize("event_cls", (WPThreadEvent, threading.Event))
def test_awareness(event_cls):

	class A:

		def __init__(self):
			self.a = True

		def __call__(self, *args, **kwargs):
			return self.a

	a = A()

	event = event_cls()
	assert(event.is_set() is False)

	assert(awareness_wait(event, a) is True)
	assert(event.is_set() is True)

	a.a = False
	assert(awareness_wait(event, a, timeout=0.5) is False)
	assert(event.is_set() is False)

	th = threading.Thread(target=lambda: awareness_wait(event, a))
	th.start()
	assert(a.a is False)
	assert(event.is_set() is False)
	event.set()
	th.join()
	assert(event.is_set() is True)
	assert(a.a is False)

	a.a = True
	assert(awareness_wait(event, a) is True)
	assert(event.is_set() is True)




from time import sleep
from random import random
from threading import Thread

import pytest
from wasp_c_extensions.cmcqueue import WCMCQueue, WCMCQueueItem


class TestWCMCQueue:

    def test_auto_ack(self):
        q = WCMCQueue()
        assert(q.messages() == 0)

        q.push(1)
        assert(q.messages() == 0)

        token1 = q.subscribe()
        assert(isinstance(token1, WCMCQueueItem) is True)
        assert(q.messages() == 1)
        assert(token1.pull() is None)
        assert(q.messages() == 1)  # there are no changes in the queue, it has the last token event

        q.push('2')
        assert(q.messages() == 2)
        assert(token1.pull() == '2')
        assert(q.messages() == 1)
        assert(token1.pull() is None)

        token2 = q.subscribe()
        assert(q.messages() == 2)

        assert(token1.pull() is None)
        assert(q.messages() == 1)  # all tokens point to the same last message

        q.push(3)
        q.push(4)
        q.push(5)
        assert(token1.pull() == 3)
        assert(token1.pull() == 4)
        assert(token2.pull() == 3)
        assert(token2.pull() == 4)
        assert(token2.pull() == 5)
        assert(token1.pull() == 5)

        token1.unsubscribe()
        assert(q.messages() == 2)  # no clean up process yet
        token2.unsubscribe()
        assert(q.messages() == 0)  # no live tokens anymore
        pytest.raises(RuntimeError, token1.unsubscribe)  # unable to unsubscribe twice
        pytest.raises(RuntimeError, token2.pull)  # unable to pull unsubscribed object

    def test_manual_ack(self):
        q = WCMCQueue(manual_acknowledge=True)
        assert(q.messages() == 0)

        q.push(1)
        assert(q.messages() == 0)

        token1 = q.subscribe()
        assert(isinstance(token1, WCMCQueueItem) is True)
        assert(q.messages() == 1)
        assert(token1.pull() is None)
        assert(q.messages() == 1)  # there are no changes in the queue, it has the last token event
        assert(token1.acknowledge() is False)

        q.push('2')
        assert(q.messages() == 2)
        assert(token1.pull() == '2')
        assert(q.messages() == 2)  # 1->2
        assert(token1.pull() == '2')  # None - > '2' it is there still

        assert(token1.acknowledge() is True)
        assert(q.messages() == 1)
        assert(token1.pull() is None)
        assert(q.messages() == 1)

        token2 = q.subscribe()
        assert(q.messages() == 2)

        assert(token1.pull() is None)
        assert(q.messages() == 2)  # ack is False because there is no user payload but internal message truncated
        assert(token1.acknowledge() is False)
        assert(q.messages() == 1)  # ack is False because there is no user payload but internal message truncated

        q.push(3)
        token1.unsubscribe()
        token2.unsubscribe()
        q.push(4)
        assert(q.messages() == 0)

    def test_item(self):
        pytest.raises(RuntimeError, WCMCQueueItem)

    def test_next(self):
        q = WCMCQueue()
        token1 = q.subscribe()
        pytest.raises(TypeError, next, token1)

        q = WCMCQueue(manual_acknowledge=True)
        token1 = q.subscribe()
        pytest.raises(StopIteration, next, token1)  # no elements to iterate

        q.push(1)
        q.push(2)
        q.push(3)

        assert(next(token1) == 1)
        assert(next(token1) == 2)
        assert(next(token1) == 3)
        pytest.raises(StopIteration, next, token1)  # no elements to iterate

        q.push(4)
        q.push(5)

        assert(next(token1) == 1)
        assert(next(token1) == 2)
        assert(next(token1) == 3)

        token1.acknowledge()  # acknowledge resets iteration
        assert(next(token1) == 2)
        assert(next(token1) == 3)
        assert(next(token1) == 4)
        assert(next(token1) == 5)
        pytest.raises(StopIteration, next, token1)  # no elements to iterate

        assert(next(token1) == 2)
        assert(next(token1) == 3)
        assert(next(token1) == 4)
        token1.unsubscribe()
        pytest.raises(RuntimeError, next, token1)  # unable to iterate unsubscribed tokens

    def test_iter(self):
        q = WCMCQueue()
        token1 = q.subscribe()
        pytest.raises(TypeError, iter, token1)

        q = WCMCQueue(manual_acknowledge=True)
        token1 = q.subscribe()
        iter(token1)  # it is ok
        assert([] == [x for x in iter(token1)])
        assert([] == [x for x in token1])

        q.push(1)
        q.push(2)
        q.push(3)

        token1.acknowledge()
        assert([2, 3] == [x for x in token1])


class TestWCMCQueueConcurrency:
    __test_running__ = False
    __threads_count__ = 50
    __input_sequence__ = [int(random() * 10) for _ in range(10 ** 4)]
    __wait_pause__ = 0.01

    class Subscriber:

        def __init__(self, queue):
            self.queue = queue
            self.token = queue.subscribe()
            self.result = []

        def thread_fn(self):
            while TestWCMCQueueConcurrency.__test_running__:
                if not self.token.has_next():
                    sleep(TestWCMCQueueConcurrency.__wait_pause__)
                item = self.token.pull()
                if item is not None:
                    self.result.append(item)

            while self.token.has_next():
                item = self.token.pull()
                if item is not None:
                    self.result.append(item)

    @pytest.mark.parametrize('runs', range(5))
    def test_concurrency(self, runs):
        subscribers = []
        queue = WCMCQueue()

        def pub_thread():
            for v in TestWCMCQueueConcurrency.__input_sequence__:
                queue.push(v)

        pub_thread = Thread(target=pub_thread)
        sub_threads = []
        for i in range(TestWCMCQueueConcurrency.__threads_count__):
            s = TestWCMCQueueConcurrency.Subscriber(queue)
            subscribers.append(s)
            sub_threads.append(Thread(target=s.thread_fn))

        TestWCMCQueueConcurrency.__test_running__ = True
        pub_thread.start()
        for th in sub_threads:
            th.start()

        pub_thread.join()
        TestWCMCQueueConcurrency.__test_running__ = False
        for th in sub_threads:
            th.join()

        for s in subscribers:
            assert(s.result == TestWCMCQueueConcurrency.__input_sequence__)




import os
import re
from setuptools import setup, Extension
from configparser import ConfigParser


__extension_section_re__ = re.compile("^extension=(.*)")

__macro_extension_re__ = re.compile("^\s*([^\s]+)(\s*=\s*([^\s]+))?")


if __name__ == "__main__":

	ext_modules = None

	config = ConfigParser()
	config.read([os.path.join(os.path.dirname(__file__), 'setup.cfg')])

	for section_name in config.sections():
		r = __extension_section_re__.search(section_name)
		if r is not None:
			section = config[section_name]
			if "sources" not in section:
				raise KeyError('"sources" must be set in "%s" section' % section_name)

			extension_kw = {}

			for option in ["include_dirs", "libraries", "extra_compile_args", "depends"]:
				if option in section:
					extension_kw[option] = section[option].split()

			if "macros" in section:
				define_macros=[]
				for single_record in section["macros"].split("\n"):

					d = __macro_extension_re__.search(single_record)
					if d is not None:
						macro_name, _, macro_value = d.groups()

						define_macros.append(
							(macro_name, macro_value if macro_value is not None else "")
						)
				extension_kw["define_macros"] = define_macros

			e = Extension(r.group(1), section["sources"].split(), **extension_kw)
			if ext_modules is None:
				ext_modules = []
			ext_modules.append(e)

	setup(ext_modules=ext_modules)
